/*
 *  ptrtd.c
 *
 *  ptrtd - Portable IPv6 TRT implementation
 *
 *  Copyright (C) 2001  Nathan Lutchansky
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

#include "config.h"
#include "event.h"
#include "defs.h"
#include "util.h"
#include "if.h"
#include "icmp.h"
#include "buffer.h"
#include "tcp.h"

void ptrtd_tcp_init(void);
void ptrtd_udp_init(void);

struct iface *iface;
char *prefix = "fec0:0:0:ffff::";
uchar pfx[16];
uchar mylladdr[16] = { 0xfe, 0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5 };
uchar otherlladdr[16] = { 0xfe, 0x80, 0, 0, 0, 0, 0, 0,
			0, 0, 0, 0, 0, 0, 0, 1 };
//uchar otherlladdr[16] = { 0xfe, 0x80, 0, 0, 0, 0, 0, 0,
//			0xfc, 0xfd, 0xc0, 0xff, 0xfe, 0xa8, 0x11, 0x02 };
char *me;
int plen = 64;
int debug = 0;

void usage(void)
{
	printf( "Usage: %s [-i [<driver>:]<interface>] [-p <prefix>] [-l <prefix length>]\n", me );
	printf( "    prefix defaults to fec0:0:0:ffff:: len 64\n" );
	exit( 1 );
}

#if 0
void dump_packet( char *title, struct pbuf *pkt )
{
	int j, len = pkt->dlen;
	char b[17];
	uchar *p = pkt->d;
	printf( "Packet: %s\n", title );

	for( j = 0; j < len; ++j )
	{
		if( j % 16 == 0 ) printf( "%08x  ", j );

		printf( "%02x ", p[j] );
		b[j % 16] = isprint( p[j] ) ? p[j] : '.';

		if( j % 16 == 15 || j == len - 1 )
		{
			b[j % 16 + 1] = 0;
			if( j == len - 1 )
			{
				j = j % 16;
				while( ++j < 16 ) printf( "   " );
				j = len;
			}
			printf( " %s\n", b );
		}
	}
}

int ip6tostr( char *str, uchar *addr )
{
	sprintf( str, "%x:%x:%x:%x:%x:%x:%x:%x",
			GET_16( addr + 0 ), GET_16( addr + 2 ),
			GET_16( addr + 4 ), GET_16( addr + 6 ),
			GET_16( addr + 8 ), GET_16( addr + 10 ),
			GET_16( addr + 12 ), GET_16( addr + 14 ) );
	return 0;
}

int strtoip6( uchar *addr, char *str )
{
	char *c;
	unsigned int tmp;
	int i = 0, skip = 0;
	char *zero = "0";
	char *g[8] = { zero, zero, zero, zero, zero, zero, zero, zero };

	for( c = str + strlen( str ); c != str; --c )
		if( c[0] == ':' )
		{
			++skip;
			if( c[-1] == ':' ) break;
		}

	if( str[0] == ':' ) c = str;
	else c = str - 1;
	for( ; c; c = strchr( c, ':' ) )
	{
		++c;
		if( *c == ':' )
		{
			i = 8 - skip;
			++c;
			if( ! *c ) break;
		}
		g[i] = c;
		++i;
	}
	for( i = 0; i < 8; ++i )
	{
		sscanf( g[i], "%x", &tmp );
		PUT_16( addr + 2 * i, tmp );
	}

	return 0;
}

int make_cksum( uchar *p, int len )
{
	int i;
	int c = GET_16( p + 4 ) + p[6];

	for( i = 0x8; i < len; ++i )
		c += (i % 2) ? p[i] : (p[i] << 8);
	while( c > 0xffff )
		c = (c >> 16) + (c & 0xffff);
	return c;
}
#endif

void handle_packet( struct iface *i, struct pbuf *p )
{
	int sum;

	sum = make_cksum( p->d, p->dlen );
	switch( p->d[6] )
	{
	case 58:
		handle_icmp( p->d, p->dlen );
		break;
	case 6:
		handle_tcp( p->d, p->dlen );
		break;
	case 17:
		handle_udp( p->d, p->dlen );
		break;
	case 0:
		break;
	default:
		icmp_send_error( p->d, p->dlen );
	}
}

void init_iface( char *type, char *dev )
{
	char cmd[256], ifname[256];
	int do_config = 0;

	iface = create_iface( type, dev, handle_packet );
	if( ! iface )
	{
		fprintf( stderr, "Unable to create %s interface.\n", type );
		exit( 1 );
	}
	if( ! strcmp( type, "tap" ) )
	{
		tap_get_name( iface, ifname );
		do_config = 1;
	}
	if( ! strcmp( type, "tun" ) )
	{
		tun_get_name( iface, ifname );
		do_config = 1;
	}
	if( do_config )
	{
		printf( "Tunnel: %s\n", ifname );
		sprintf( cmd, "/sbin/ip link set %s up", ifname );
		printf( "command: %s\n", cmd );
		system( cmd );
		sprintf( cmd, "/sbin/ip addr add fe80::1/64 dev %s", ifname );
		printf( "command: %s\n", cmd );
		system( cmd );
		sprintf( cmd, "/sbin/ip route add %s/%d dev %s via fe80::5",
				prefix, plen, ifname );
		printf( "command: %s\n", cmd );
		system( cmd );
	}
	icmp_init_iface( iface );
}

int main( int argc, char **argv )
{
	int c;
	char *itype = "tap", *iname = NULL;
	char temp[40];

	me = argv[0];

	setvbuf( stdout, NULL, _IONBF, 0 );

	init_iface_drivers();

	while( ( c = getopt( argc, argv, "di:l:p:" ) ) != -1 )
	{
		switch( c )
		{
		case 'd':
			debug = 1;
			break;
		case 'i':
			if( ( iname = strchr( optarg, ':' ) ) )
			{
				*iname = 0;
				++iname;
				itype = optarg;
			} else iname = optarg;
			break;
		case 'l':
			plen = atoi( optarg );
			break;
		case 'p':
			prefix = optarg;
			break;
		default:
			usage();
			break;
		}
	}

	strtoip6( pfx, prefix );
	ip6tostr( temp, pfx );
	printf( "using prefix %s\n", temp );

	ptrtd_tcp_init();
	ptrtd_udp_init();

	init_iface( itype, iname );

	if( ! debug )
		if( daemon(0,0) < 0 )
		{
			perror( "daemon" );
			exit( 1 );
		}

	event_loop();
	return 0;
}
